#ifndef _GRECS_JSONPARSE_H
#define _GRECS_JSONPARSE_H

#include <stddef.h>
#include "grecs/json.h"
#include "grecs/locus.h"

struct json_value *grecs_json_parse_string(char const *input, size_t length,
					   char const **errmsg,
					   struct grecs_locus *errloc);
struct json_value *json_value_lookup(struct json_value *obj,
				     const char *ident);
#endif

