/* A Bison parser, made by GNU Bison 3.8.2.  */

/* Bison interface for Yacc-like parsers in C

   Copyright (C) 1984, 1989-1990, 2000-2015, 2018-2021 Free Software Foundation,
   Inc.

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */

/* As a special exception, you may create a larger work that contains
   part or all of the Bison parser skeleton and distribute that work
   under terms of your choice, so long as that work isn't itself a
   parser generator using the skeleton or a modified version thereof
   as a parser skeleton.  Alternatively, if you modify or redistribute
   the parser skeleton itself, you may (at your option) remove this
   special exception, which will cause the skeleton and the resulting
   Bison output files to be licensed under the GNU General Public
   License without this special exception.

   This special exception was added by the Free Software Foundation in
   version 2.2 of Bison.  */

/* DO NOT RELY ON FEATURES THAT ARE NOT DOCUMENTED in the manual,
   especially those whose name start with YY_ or yy_.  They are
   private implementation details that can be changed or removed.  */

#ifndef YY_GRECS_META1_META1_GRAM_H_INCLUDED
# define YY_GRECS_META1_META1_GRAM_H_INCLUDED
/* Debug traces.  */
#ifndef GRECS_META1_DEBUG
# if defined YYDEBUG
#if YYDEBUG
#   define GRECS_META1_DEBUG 1
#  else
#   define GRECS_META1_DEBUG 0
#  endif
# else /* ! defined YYDEBUG */
#  define GRECS_META1_DEBUG 1
# endif /* ! defined YYDEBUG */
#endif  /* ! defined GRECS_META1_DEBUG */
#if GRECS_META1_DEBUG
extern int grecs_meta1_debug;
#endif
/* "%code requires" blocks.  */
#line 31 "meta1-gram.y"

typedef void *yyscan_t;
struct parser_control {
	struct grecs_node *parse_tree;
	struct grecs_locus_point point;
	struct grecs_txtacc *acc;
};
#define GRECS_META1_LTYPE grecs_locus_t

#line 67 "meta1-gram.h"

/* Token kinds.  */
#ifndef GRECS_META1_TOKENTYPE
# define GRECS_META1_TOKENTYPE
  enum grecs_meta1_tokentype
  {
    GRECS_META1_EMPTY = -2,
    GRECS_META1_EOF = 0,           /* "end of file"  */
    GRECS_META1_error = 256,       /* error  */
    GRECS_META1_UNDEF = 257,       /* "invalid token"  */
    META1_STRING = 258,            /* META1_STRING  */
    META1_IDENT = 259              /* META1_IDENT  */
  };
  typedef enum grecs_meta1_tokentype grecs_meta1_token_kind_t;
#endif

/* Value type.  */
#if ! defined GRECS_META1_STYPE && ! defined GRECS_META1_STYPE_IS_DECLARED
union GRECS_META1_STYPE
{
#line 55 "meta1-gram.y"

	char *string;
	grecs_value_t svalue, *pvalue;
	struct grecs_list *list;
	struct { struct grecs_node *head, *tail; } node_list;
	struct grecs_node *node;

#line 96 "meta1-gram.h"

};
typedef union GRECS_META1_STYPE GRECS_META1_STYPE;
# define GRECS_META1_STYPE_IS_TRIVIAL 1
# define GRECS_META1_STYPE_IS_DECLARED 1
#endif

/* Location type.  */
#if ! defined GRECS_META1_LTYPE && ! defined GRECS_META1_LTYPE_IS_DECLARED
typedef struct GRECS_META1_LTYPE GRECS_META1_LTYPE;
struct GRECS_META1_LTYPE
{
  int first_line;
  int first_column;
  int last_line;
  int last_column;
};
# define GRECS_META1_LTYPE_IS_DECLARED 1
# define GRECS_META1_LTYPE_IS_TRIVIAL 1
#endif




int grecs_meta1_parse (struct parser_control *pctl, void *yyscanner);

/* "%code provides" blocks.  */
#line 40 "meta1-gram.y"

int grecs_meta1_lex(GRECS_META1_STYPE *lvalp, GRECS_META1_LTYPE *llocp,
		    yyscan_t yyscanner);

struct grecs_node *grecs_meta1_parser(const char *name, int traceflags);
yyscan_t grecs_meta1_new_scanner(struct parser_control *pctl, int debug);
int grecs_meta1_new_source(yyscan_t scanner, const char *name);
void grecs_meta1_close_scanner(yyscan_t scanner);
void grecs_meta1_error (GRECS_META1_LTYPE const *llocp,
			struct parser_control *pctl, yyscan_t scanner,
			char const *msg);

#line 137 "meta1-gram.h"

#endif /* !YY_GRECS_META1_META1_GRAM_H_INCLUDED  */
